/*
 * This is FXcube, a tiny Linux demo where I collected my favourite
 * oldskool effects on a single 3d spinning cube! Have fun. To skip
 * scenes, press keys. Some 100 MHz needed. You need X11 with 16bit
 * colors and 1.5 MB of free RAM, plus /dev/dsp for fake music :-).
 * Copyright 1994-2003 by Eric Auer <eric@coli.uni-sb.de>.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should find a copy of the GNU General Public License on the
 * site where you have downloaded this source. If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307 USA
 * (or try http://www.gnu.org/licenses/licenses.html at www.gnu.org).
 */

/* Based on textur3.pas, AuerSoft -16.12.1995 */

#include "textur.h"

/*
 * given screen space origin, and target, set up screen space.
 */
void setview(const VEC * e0, const VEC *target, space3d * scr3d)
{
  VEC yaxis  = { 0, 1, 0 }; /* y unit axis */

    /* set up screen space origin */
  vCOPYv(scr3d->e0, *e0);
    /* z axis is view direction: */
  vCOPYv(scr3d->ez0, *target);
  vSUBv(scr3d->ez0, scr3d->e0);
  vMULs(scr3d->ez0, 1.0 / vABS(scr3d->ez0)); /* scale to length 1 */
    /* x axis is ortho to view, y unit: */
  vCROSSv(yaxis, scr3d->ez0, scr3d->ex0); 
  vMULs(scr3d->ex0, 1.0 / vABS(scr3d->ex0)); /* scale to length 1 */
    /* y axis is ortho to x and z axis: */
  vCROSSv(scr3d->ez0, scr3d->ex0, scr3d->ey0);
  vMULs(scr3d->ey0, 1.0 / vABS(scr3d->ey0)); /* scale to length 1 */
} /* setview */


/*
 * do 3d to 2d projection of a single point
 * given screen and a 3d point, find a 2d point.
 */
void bildpunkt(const space3d * screen,
  const VEC * iPt, pointint * oiPt, V_FP * dist)
{
  V_FP rg;     /* intermediate result */
  pointfp oPt; /* projected point */
  VEC distv;    /* p0 - e0 */
  /* xe = (distv.x * scrdist) / (distv.z + scrdist); */
  /* ye = (distv.y * scrdist) / (distv.z + scrdist); */
  /* 3/2003: NOT (distv.... * distv.z) / (distv.z * scrdist) !!! */
  
  vCOPYv(distv, *iPt);
  vSUBv(distv, screen->e0);
  rg = vMULv(distv, screen->ez0);

#if PERSPECTIVE
  /* perspective projection, assuming a screen glass plane  */
  /* between the eye-point and the point. The eye-point is  */
  /* scrdist units from e0 of the screen, in -ez0 direction */
  if (rg < 0) {
    /* point is on the user side of the screen glass plane...    */
    /* could change color here or set a return value or similar. */
    if ((rg + scrdist) > (scrdist / 2)) {
      oPt.x = scrdist * vMULv(distv, screen->ex0) / (rg + scrdist);
      oPt.y = scrdist * vMULv(distv, screen->ey0) / (rg + scrdist);
    } else {
      /* point is so close that coordinates would be bigger than */
      /* two times the distv.x and distv.y values, reject it.    */
      oPt.x = 0;
      oPt.y = 0;
    }
  } else {
    oPt.x = scrdist * vMULv(distv, screen->ex0) / (rg + scrdist);
    oPt.y = scrdist * vMULv(distv, screen->ey0) / (rg + scrdist);
  }
  oiPt->x = round(oPt.x * skalx) + xm;
  oiPt->y = round(oPt.y * skaly) + ym;
#else
  /* rectangular (non-perspective) projection: */
  oPt.x = vMULv(distv, screen->ex0);
  oPt.y = vMULv(distv, screen->ey0);
  oiPt->x = round(oPt.x * RECTFACTOR * skalx) + xm;
  oiPt->y = round(oPt.y * RECTFACTOR * skaly) + ym;
#endif


  *dist = rg + scrdist;

} /* bildpunkt */


/*
 * given the screen space and a 3d plane, calculate a texture map.
 * We transform the plane data to screen relative format and then
 * store the result in a single structure instead of 3 vectors.
 */
void settextur(const space3d * screen,
  const plane3d * plane, texturemap * tmap)
{
  VEC dist; /* p0 - e0 */
  
  vCOPYv(dist, plane->p0);
  vSUBv(dist, screen->e0);
  /* could normalize px0 py0 to unit length here */
  tmap->x0 = vMULv(screen->ex0, dist);
  tmap->y0 = vMULv(screen->ey0, dist);
  tmap->z0 = vMULv(screen->ez0, dist);
  tmap->xx = vMULv(screen->ex0, plane->px0);
  tmap->yx = vMULv(screen->ey0, plane->px0);
  tmap->zx = vMULv(screen->ez0, plane->px0);
  tmap->xy = vMULv(screen->ex0, plane->py0);
  tmap->yy = vMULv(screen->ey0, plane->py0);
  tmap->zy = vMULv(screen->ez0, plane->py0);
} /* settextur */


/*
 * given 2d coordinates on a mapped texture, find screen coordinates.
 * Must be consistent to bildpunkt() !!!
 */
void bildtextur(const pointint * iPt, const texturemap * tmap,
  pointint * oiPt)
{
  pointfp oPt;  /* destination point  */

#if PERSPECTIVE
  V_FP rg;      /* intermediate value */

  rg = tmap->z0 + (iPt->x * tmap->zx)
     + (iPt->y * tmap->zy);
    /* 0 if iPt in screen glass plane */

  /* bildpunkt() correction 3/2003: factor is scrdist, not rg! */
  oPt.x = scrdist
    * (tmap->x0 + (iPt->x * tmap->xx) + (iPt->y * tmap->xy))
    / (rg + scrdist);
  oPt.y = scrdist
    * (tmap->y0 + (iPt->x * tmap->yx) + (iPt->y * tmap->yy))
    / (rg + scrdist);
  oiPt->x = round(oPt.x * skalx) + xm;
  oiPt->y = round(oPt.y * skaly) + ym;
#else
  oPt.x = (tmap->x0 + (iPt->x * tmap->xx) + (iPt->y * tmap->xy));
  oPt.y = (tmap->y0 + (iPt->x * tmap->yx) + (iPt->y * tmap->yy));
  oiPt->x = round(oPt.x * RECTFACTOR * skalx) + xm;
  oiPt->y = round(oPt.y * RECTFACTOR * skaly) + ym;
#endif

} /* bildtextur */


/*
 * rotate a 3d point around the y axis around the origin
 * iPt, oPt must be different.
 * Strange: xxx->x = yyy->y destroys things!
 */
void rotiere_y(int angle, const VEC * iPt, VEC * oPt)
{
  V_FP co, si;

  /* could subtract center before */
  co = cos((PI * angle) / 180);
  si = sin((PI * angle) / 180);

  oPt->x =  (iPt->x * co) + (iPt->z * si);
  oPt->y =  iPt->y;
  oPt->z = -(iPt->x * si) + (iPt->z * co);
  /* could add center after */
    
} /* rotiere_y */


/*
 * rotate a 3d point around the z axis around the origin
 * iPt, oPt must be different.
 */
void rotiere_z(int angle, const VEC * iPt, VEC * oPt)
{
  V_FP co, si;

  /* could subtract center before */
  co = cos((PI * angle) / 180);
  si = sin((PI * angle) / 180);

  oPt->x =  (iPt->x * co) + (iPt->y * si);
  oPt->y = -(iPt->x * si) + (iPt->y * co);
  oPt->z =  iPt->z;
  /* could add center after */

} /* rotiere_z */


/*
 * given a screen space,
 * 3 indexes (org, x, y), an array of 3d points,
 * screen space, the resolution, a texture buffer,
 * the x y texture origin, a destination buffer,
 * map the texture in a perspective (or polygon) way.
 */
void cubeplane(const space3d * scr3d, int oidx, int xidx, int yidx,
  VEC * pts3d, int resx, int resy,
  const EA_PIXEL * inbuf, int xorg, int yorg,
  EA_PIXEL * outbuf)
{
  plane3d plane;   /* source plane */      /* p0 px0 py0 */
  texturemap tmap; /* texture map precalculated values */
  pointint srcIpt; /* source point */      /* xp yp */
  pointint dstIpt; /* destination point */ /* xei yei */
#if FULLPERSPECTIVE
#else
  /* poly mapping style: use 16.16 fixed point variables! */
  pointint corners[4];  /* poly corners */ /* TL, BL, TR, TB */
  pointint leftIpt;     /* incremented as Y changes */
  pointint dx, dy;      /* projected increments */
  pointint ddx;         /* increment on X increment over Y */
  int doublex, doubley; /* pixel doubling */
#endif

  vCOPYv(plane.p0,  pts3d[oidx]);

  vCOPYv(plane.px0, pts3d[xidx]);
  vSUBv(plane.px0,  pts3d[oidx]);
  vMULs(plane.px0,  1.0 / (resx-1));

  vCOPYv(plane.py0, pts3d[yidx]);
  vSUBv(plane.py0,  pts3d[oidx]);
  vMULs(plane.py0,  1.0 / (resy-1));

#if FULLPERSPECTIVE
  settextur(scr3d, &plane, &tmap);
#else
  settextur(scr3d, &plane, &tmap);
  srcIpt.x = 0;
  srcIpt.y = 0;
  bildtextur(&srcIpt, &tmap, &(corners[0]));
  srcIpt.x = resx - 1;
  bildtextur(&srcIpt, &tmap, &(corners[1]));
  srcIpt.y = resy - 1;
  bildtextur(&srcIpt, &tmap, &(corners[3]));
  srcIpt.x = 0;
  bildtextur(&srcIpt, &tmap, &(corners[2]));
  v2MULs(corners[0], 1<<16); /* fixed point scaling */
  v2MULs(corners[1], 1<<16); /* fixed point scaling */
  v2MULs(corners[2], 1<<16); /* fixed point scaling */
  v2MULs(corners[3], 1<<16); /* fixed point scaling */
  v2COPYv(leftIpt, corners[0]); /* start of current line */
  v2COPYv(dx, corners[1]);
  v2SUBv(dx, corners[0]);
  dx.x /= (resx-1); /* dx of first line */
  dx.y /= (resx-1); /* dx of first line */
  v2COPYv(ddx, corners[3]);
  v2SUBv(ddx, corners[2]);
  ddx.x /= (resx-1); /* dx of last line */
  ddx.y /= (resx-1); /* dx of last line */
  v2SUBv(ddx, dx); /* diff between dx in first and last line */
  ddx.x /= (resy-1); /* delta between deltas */
  ddx.y /= (resy-1); /* delta between deltas */
  v2COPYv(dy, corners[2]);
  v2SUBv(dy, corners[0]);
  dy.x /= (resy-1); /* delta between line starts */
  dy.y /= (resy-1); /* delta between line starts */
#endif

        /* could skip copying if source pixel is black   */
        /* (but then we would have to draw the cube back */
        /* planes as well, which doubles CPU usage...).  */

#if FULLPERSPECTIVE

  for (srcIpt.x=0; srcIpt.x<resx; srcIpt.x++) {
    for (srcIpt.y=0; srcIpt.y<resy; srcIpt.y++) {
      /* for testing: dstIpt = srcIpt; */

      /* inlined call of: bildtextur(&srcIpt, &tmap, &dstIpt); */
      V_FP rg;      /* intermediate value */
      pointfp oPt;  /* destination point  */
      rg = tmap.z0 + (srcIpt.x * tmap.zx) + (srcIpt.y * tmap.zy);
      oPt.x = rg * (tmap.x0 + (srcIpt.x * tmap.xx)
        + (srcIpt.y * tmap.xy)) / (rg + scrdist);
      dstIpt.x = round(oPt.x * skalx) + xm;
      oPt.y = rg * (tmap.y0 + (srcIpt.x * tmap.yx)
        + (srcIpt.y * tmap.yy)) / (rg + scrdist);
      dstIpt.y = round(oPt.y * skaly) + ym;
      /* end of inlined bildtextur() call */
#if CLIPTEXTURE
      if ((dstIpt.x >= 0) && (dstIpt.x < (W-1)) &&
        (dstIpt.y >= 0) && (dstIpt.y < (H-1)))
#endif
        if (skalx > resx) {
          GFX(dstIpt.x, dstIpt.y, outbuf) =
          GFX(dstIpt.x + 1, dstIpt.y, outbuf) =
            GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
        } else {
          GFX(dstIpt.x, dstIpt.y, outbuf) =
            GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
        }
        if (skaly > resy) {
          if (skalx > resx) {
            GFX(dstIpt.x, dstIpt.y + 1, outbuf) =
            GFX(dstIpt.x + 1, dstIpt.y + 1, outbuf) =
              GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
          } else {
            GFX(dstIpt.x, dstIpt.y + 1, outbuf) =
              GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
          }
        }
    }
  }

#else

  for (srcIpt.y=0; srcIpt.y<resy; srcIpt.y++) {
    v2COPYv(dstIpt, leftIpt); /* start of line */
    doublex = ( (abs(dx.x) + abs(dy.x)) > (1<<16));
    doubley = ( (abs(dx.y) + abs(dy.y)) > (1<<16));
    for (srcIpt.x=0; srcIpt.x<resx; srcIpt.x++) {
#if CLIPTEXTURE
      if ((dstIpt.x >= 0) && (dstIpt.x < ((W-1) << 16)) &&
        (dstIpt.y >= 0) && (dstIpt.y < ((H-1) << 16))) {
#else
      {
#endif
        if (doublex) {
          GFX((dstIpt.x >> 16), (dstIpt.y >> 16), outbuf) =
          GFX((dstIpt.x >> 16) + 1, (dstIpt.y >> 16), outbuf) =
            GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
        } else {
          GFX((dstIpt.x >> 16), (dstIpt.y >> 16), outbuf) =
            GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
        }
        if (doubley) {
          if (doublex) {
            GFX((dstIpt.x >> 16), (dstIpt.y >> 16) + 1, outbuf) =
            GFX((dstIpt.x >> 16) + 1, (dstIpt.y >> 16) + 1, outbuf) =
              GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
          } else {
            GFX((dstIpt.x >> 16), (dstIpt.y >> 16) + 1, outbuf) =
              GFX(xorg + srcIpt.x, yorg + srcIpt.y, inbuf);
          }
        }
      } /* clipping check */
      v2ADDv(dstIpt, dx); /* traverse this line further */
    }
    v2ADDv(dx, ddx); /* update X increment for new line */
    v2ADDv(leftIpt, dy);       /* jump to the next line */
  }

#endif

} /* cubeplane */


/*
 * draw a complete 3d cube, rotated by given angles, using textures
 * from one buffer, to another buffer.
 */
void cube(int phi1, int phi2, const EA_PIXEL * inbuf, EA_PIXEL * outbuf)
{
  space3d scr3d;   /* screen space */      /* e0 ex0 ey0 ey0 */

  VEC target =    { 0.5,   0, 2 }; /* where to look */
  VEC center =    { 0  , 0.5, 2 }; /* center of the cube */
  VEC midscreen = { 0  ,   2, 0 }; /* origin of the screen space */
  int texResX = W>>2;          /* texture width and height */
  int texResY = H>>1;          /* texture width and height */
  V_FP phi3;

  VEC p0;
  VEC p1;
  V_FP min;
  pointint ip0;

  VEC pts3d[8];     /* cube corners, 3d */
  pointint ipts[8]; /* cube corners, 2d projected, int */
  V_FP dists[6];    /* corner distances */

  int i,j;          /* counter, temp value */
  int planes[6];    /* distance order of planes */
  V_FP pdists[6];   /* plane distances */
  V_FP cosines[6];  /* cosines of plane viewing angles */
  
  setview(&midscreen, &target, &scr3d);
  
  /*     6---7       Segments:
   *    /!  /!       A B C D
   *   2===3 !       E F G H
   *   I 4-I-5
   *   I/  I/
   *   0===1
   */

#ifdef DEBUG   
  printf("Cube projection: phi1=%d, phi2=%d\n", phi1, phi2);
#endif

  /* derived phi3 motion, for extra seasickness! */
  phi3 = PI * (phi1 + phi2) / 180;
  if ( phi3 > ( PI + PI ) )
    phi3 = phi3 - ( PI + PI );
  center.x = center.x + ( 0.125 * sin(phi3) );
  center.y = center.y + ( 0.125 * cos(phi3) );
  center.z = center.z + ( 0.25 * sin(phi3) );

  for (i=0; i<8; i++) {
    p0.x = ((i & 1) > 0) ? 0.5 : -0.5;
    p0.y = ((i & 2) > 0) ? 0.5 : -0.5;
    p0.z = ((i & 4) > 0) ? 0.5 : -0.5;

    rotiere_y(phi1, &p0, &p1);
    rotiere_z(phi2, &p1, &(pts3d[i]));

    vADDv(pts3d[i], center);
    
    bildpunkt(&scr3d, &(pts3d[i]), &(ipts[i]), &(dists[i]));

#ifdef DEBUG3D
    printf("pts3d[%d] = { %f %f %f } ipts[%d] = { %d %d } dist = %f\n",
      i, pts3d[i].x, pts3d[i].y, pts3d[i].z,
      i, ipts[i].x, ipts[i].y, dists[i]);
#endif
  }

  /* measure plane distances AND relative angles */
  /* if cosine of angle is negative, plane is "visible" from  */
  /* the back side. If the cube is not transparent, this only */
  /* happens for hidden planes. Angles are also useful for    */
  /* possible lighting, but then we would use light vector-   */
  /* relative angles...                                       */

#define DETERMINE(a, b, c, i) \
  p0.x = 0.5 * (a);   p0.y = 0.5 * (b);   p0.z = 0.5 * (c); \
  rotiere_y(phi1, &p0, &p1); \
  rotiere_z(phi2, &p1, &p0); \
  vADDv(p0, center); /* get plane center point */ \
  bildpunkt(&scr3d, &p0, &ip0, &(pdists[(i)])); /* find distance */ \
  p0.x = (a);   p0.y = (b);  p0.z = (c); \
  rotiere_y(phi1, &p0, &p1); \
  rotiere_z(phi2, &p1, &p0); \
  /* now we have an unit length plane orthogonal vector */ \
  cosines[(i)] = -vMULv(p0, scr3d.ez0) \
  /* get cosine of angle to space Z */
  /* PROBLEM:  things end up > unit length, no real cosine (why?) */
  
  DETERMINE( 0,  1,  0, 0); /* top    plane */
  DETERMINE( 0, -1,  0, 1); /* bottom plane */
  DETERMINE( 0,  0, -1, 2); /* front  plane */
  DETERMINE( 1,  0,  0, 3); /* right  plane */
  DETERMINE( 0,  0,  1, 4); /* back   plane */
  DETERMINE(-1,  0,  0, 5); /* left   plane */

#ifdef DEBUG
  printf("Plane distances: ");
  for (i=0; i<6; i++)
    printf("[%d]=%2.2f ", i, pdists[i]);
  printf("\n");
  printf("Angle cosines:   ");
  for (i=0; i<6; i++)
    printf("[%d]=% 1.2f ", i, cosines[i]);
  printf("\n");
#endif

#define MINCOS 0.0 /* 0 to ignore plane backsides, -1.1 otherwise */

  /* not using qsort and stdlib, just find closest 3 planes   */
  /* cleaner solution would be having an array of cubeplane() */
  /* calls and insertion-sorting it on creation...            */
  /* (as in for_3d-using-asm-tsr.pas.gz)                      */
  
  for (i=0; i<6; i++)
    planes[i] = -1;

  min = 123456.78;
  for (i=0; i<6; i++) {
    if ((pdists[i] < min)
      && (cosines[i] > MINCOS)) {
      min = pdists[i];
      planes[0] = i; /* closest */
    }
  }
  min = 123456.78;
  for (i=0; i<6; i++) {
    if ((pdists[i] < min) && (cosines[i] > MINCOS)
      && (i != planes[0])) {
      min = pdists[i];
      planes[1] = i; /* second closest */
    }
  }
  min = 123456.78;
  for (i=0; i<6; i++) {
    if ((pdists[i] < min)
      && (cosines[i] > MINCOS)
      && (i != planes[0]) && (i != planes[1])) {
      min = pdists[i];
      planes[2] = i; /* third closest */
    }
  }
  /* we do not consider planes that are further away */
  
  min = -1.0; /* worst case: parallel to screen, pointing to back */
  j = -1; /* no best match found yet */
  for (i=0; i<6; i++) {
    if (cosines[i] > min) {
      j = i; /* relatively "parallel to screen / points to front" */
      min = cosines[i];
    }
  }

  if (planes[0] != j)
    printf("Draw: %d, %d, %d. Most visible: %d !?!?\n",
      planes[2], planes[1], planes[0], j);
#ifdef DEBUG
  else
    printf("Draw: %d, %d, %d.\n", planes[2], planes[1], planes[0]);
#endif

  /* you could actually draw all 3 planes which are closer than */
  /* the cube center in any order, as they should never be in   */
  /* the way of each other. This would remove the need to sort. */
  /* WE DRAW FROM FAR TO NEAR, OF COURSE. */
  for (i=2; i>=0; i--) {  /* only paint front 3 planes */
    switch (planes[i]) { /* select and draw one plane */
      case 0:
      /* top    plane: 2, 3, 7, 6, taken from segment C */
      cubeplane(&scr3d, 2, 3, 6, pts3d, texResX, texResY,
        inbuf, W>>1, 0, outbuf); break;
      case 1:
      /* bottom plane: 0, 4, 5, 1, taken from segment D */
      cubeplane(&scr3d, 0, 4, 1, pts3d, texResX, texResY,
        inbuf, 3*(W>>2), 0, outbuf); break;
      case 2:
      /* front plane: 0, 1, 3, 2, taken from segment E */
      cubeplane(&scr3d, 0, 1, 2, pts3d, texResX, texResY,
        inbuf, 0, H>>1, outbuf); break;
      case 3:
      /* right plane: 1, 5, 7, 3, taken from segment F */
      cubeplane(&scr3d, 1, 5, 3, pts3d, texResX, texResY,
        inbuf, W>>2, H>>1, outbuf); break;
      case 4:
      /* back  plane: 5, 4, 6, 7, taken from segment G */
      cubeplane(&scr3d, 5, 4, 7, pts3d, texResX, texResY,
        inbuf, W>>1, H>>1, outbuf); break;
      case 5:
      /* left  plane: 4, 0, 2, 6, taken from segment H */
      cubeplane(&scr3d, 4, 0, 6, pts3d, texResX, texResY,
        inbuf, 3*(W>>2), H>>1, outbuf); break;
    } /* switch */
  } /* for */

} /* cube */

